@@ -181,6 +181,22 @@
 //!
 //! assert_eq!(results, vec![Ok(0), Ok(1), Ok(2), Err(()), Ok(4)])
 //! ```
+//! ## Overhead Benchmark
+//! Platform: Macbook Air (2015 Late) 8 GB RAM, Intel Core i5, 1.6GHZ (2 Core).
+//!
+//! ### Result
+//! One million (1,000,000) empty iteration for each run.
+//! ```text
+//! test iter_async::test_par_iter_async::bench_into_par_iter_async
+//!     ... bench: 110,277,577 ns/iter (+/- 28,510,054)
+//!
+//! test test_par_iter::bench_into_par_iter_sync
+//!     ... bench: 121,063,787 ns/iter (+/- 103,787,056)
+//! ```
+//!
+//! Result:
+//! - Async iterator overhead `110 ns (+/-  28 ns)`.
+//! - Sync iterator overhead  `121 ns (+/- 103 ns)`.
 //!
 //! ## Implementation Note
 //!
@@ -190,39 +206,38 @@
 //!   get blocked. The channel size is hard-coded to 100 for each thread.
 //! - The number of threads equals to the number of logical cores.
 //!
-//! ### Synchronization Mechanism
-//! - When each thread fetch a task, it registers its thread ID (`thread_number`)
-//!   and the task ID (`task_number`) into a mpsc channel.
-//! - When `next()` is called, the consumer fetch from the task registry
-//!   (`task_order`) the next thread ID and task ID.
-//! - If `next()` detect that some thread has not produced result due to exception,
-//!   it calls `kill()`, which stop threads from fetching new tasks,
-//!   flush remaining tasks, and joining the worker threads.
-//!
-//! ### Error handling and Dropping
-//! - When any exception occurs, stop producers from fetching new task.
-//! - Before dropping the structure, stop all producers from fetching tasks,
-//!   flush all remaining tasks, and join all threads..
+//! ### Synchronization and Exception Handling
+//! - When each thread fetch a task, it registers its thread ID and task ID into a registry.
+//! - When `next()` is called, the consumer fetch from the task registry the next thread ID.
+//! - `next()` returns None if there is no more task or if some Error occurs.
 //!
 mod iter_async;
 
-use crossbeam::channel::{bounded, unbounded, Receiver, Sender};
+use crossbeam::channel::{bounded, Receiver};
+use crossbeam::sync::{Parker, Unparker};
+use crossbeam::utils::Backoff;
 pub use iter_async::*;
 use num_cpus;
-use std::sync::atomic::{AtomicBool, Ordering};
-use std::sync::{Arc, Mutex};
+use std::ops::Deref;
+use std::sync::atomic::{AtomicBool, AtomicIsize, Ordering};
+use std::sync::Arc;
 use std::thread;
 use std::thread::JoinHandle;
+use std::time::Duration;
 
-const MAX_SIZE_FOR_THREAD: usize = 100;
+const MAX_SIZE_FOR_THREAD: usize = 128;
+const BUFFER_SIZE: usize = 64;
 
+///
+/// lock-free sequential parallel iterator
+///
 pub trait IntoParallelIteratorSync<R, T, TL, F>
-where
-    F: Send + Clone + 'static + Fn(T) -> Result<R, ()>,
-    T: Send,
-    TL: Send + IntoIterator<Item = T>,
-    <TL as IntoIterator>::IntoIter: Send + 'static,
-    R: Send,
+    where
+        F: Send + Clone + 'static + Fn(T) -> Result<R, ()>,
+        T: Send + 'static,
+        TL: Send + IntoIterator<Item = T> + 'static,
+        <TL as IntoIterator>::IntoIter: Send + 'static,
+        R: Send,
 {
     ///
     /// # Usage
@@ -271,74 +286,239 @@ where
 }
 
 impl<R, T, TL, F> IntoParallelIteratorSync<R, T, TL, F> for TL
-where
-    F: Send + Clone + 'static + Fn(T) -> Result<R, ()>,
-    T: Send,
-    TL: Send + IntoIterator<Item = T>,
-    <TL as IntoIterator>::IntoIter: Send + 'static,
-    R: Send + 'static,
+    where
+        F: Send + Clone + 'static + Fn(T) -> Result<R, ()>,
+        T: Send + 'static,
+        TL: Send + IntoIterator<Item = T> + 'static,
+        <TL as IntoIterator>::IntoIter: Send + 'static,
+        R: Send + 'static,
 {
     fn into_par_iter_sync(self, func: F) -> ParIterSync<R> {
         ParIterSync::new(self, func)
     }
 }
 
-/// iterate through blocks according to array index.
+///
+/// A lookup table to register and look up corresponding thread id for a task.
+///
+/// `-1` represents the task ID is not yet registered
+/// or the task does not exist.
+///
+struct TaskRegistry {
+    // vector of thread ID
+    inner: Arc<Vec<AtomicIsize>>,
+    parkers: Vec<Parker>,
+}
+
+impl Deref for TaskRegistry {
+    type Target = Vec<AtomicIsize>;
+
+    fn deref(&self) -> &Self::Target {
+        self.inner.deref()
+    }
+}
+
+/// Write client for Task registry
+struct TaskRegistryWrite {
+    inner: Arc<Vec<AtomicIsize>>,
+    unparkers: Vec<Unparker>
+}
+
+impl Deref for TaskRegistryWrite {
+    type Target = Vec<AtomicIsize>;
+
+    fn deref(&self) -> &Self::Target {
+        self.inner.deref()
+    }
+}
+
+impl Drop for TaskRegistryWrite {
+    fn drop(&mut self) {
+        for unparker in &self.unparkers {
+            unparker.unpark();
+        }
+    }
+}
+
+impl TaskRegistry {
+
+    ///
+    /// Initialize the registry with `-1` to represent an empty registry
+    ///
+    /// The `size` must be (just) big enough to ensure that no key collision would
+    /// possibly occur.
+    ///
+    fn new(size: usize) -> TaskRegistry {
+        TaskRegistry {
+            inner: Arc::new((0..size).map(|_| AtomicIsize::new(-1)).collect()),
+            parkers: (0..size).map(|_| Parker::new()).collect()
+        }
+    }
+
+    ///
+    /// Look up a thread_number of a task and set that slot to `-1`.
+    ///
+    /// This function blocks to wait for a task to be registered,
+    /// unless all worker threads have stopped so that no more new
+    /// task can possibly be registered.
+    ///
+    /// It should block very rarely since task dispatcher is not blocking,
+    /// and is registered immediately after fetching in `get_task`.
+    ///
+    /// returns `None` only when all worker threads have stopped
+    ///
+    #[inline(always)]
+    pub(crate) fn lookup(&self, task_id: usize) -> Option<isize> {
+        let registry_len = self.len();
+        let pos = TaskRegistry::id_to_key(task_id, registry_len);
+        let backoff = Backoff::new();
+        loop {
+            // check if worker threads are still active
+            if !self.is_disconnected() {
+                let thread_num = self[pos].swap(-1, Ordering::SeqCst);
+                // if `-1` is read, would continue in the loop
+                if thread_num >= 0 {
+                    return Some(thread_num);
+                } else {
+                    // snooze
+                    if backoff.is_completed() {
+                        // park but no more than 500 millis
+                        self.parkers[pos].park_timeout(Duration::from_millis(500));
+                    } else {
+                        backoff.snooze();
+                    }
+                }
+                // if worker threads are no more active, might return `None`
+            } else {
+                let thread_num = self[pos].swap(-1, Ordering::SeqCst);
+                return if thread_num >= 0 {
+                    Some(thread_num)
+                } else {
+                    None
+                };
+            }
+        }
+    }
+
+    /// key of task ID in registry
+    #[inline(always)]
+    fn id_to_key(task_id: usize, registry_len: usize) -> usize {
+        task_id % registry_len
+    }
+
+    fn to_write(&self) -> TaskRegistryWrite {
+        TaskRegistryWrite {
+            inner: self.inner.clone(),
+            unparkers: self.parkers.iter().map(|p| p.unparker().clone()).collect(),
+        }
+    }
+
+    #[inline(always)]
+    fn is_disconnected(&self) -> bool {
+        Arc::strong_count(&self.inner) == 1
+    }
+}
+
+impl TaskRegistryWrite {
+    ///
+    /// When a worker fetches a new task, it calls `register` to tell
+    /// the user thread its own thread ID.
+    ///
+    /// register the worker thread number of a task
+    ///
+    #[inline(always)]
+    pub(crate) fn register(&self, task_id: usize, thread_id: isize) {
+        let registry_len = self.len();
+        let key = TaskRegistry::id_to_key(task_id, registry_len);
+        // never overwrite
+        debug_assert_eq!(self[key].load(Ordering::SeqCst), -1);
+        self[key].store(thread_id, Ordering::SeqCst);
+        self.unparkers[key].unpark();
+    }
+}
+
+///
+/// implementation of lock-free sequential parallel iterator
+///
 pub struct ParIterSync<R> {
+
     /// Result receivers, one for each worker thread
-    receivers: Vec<Receiver<R>>,
-    /// Receiver<thread>
-    task_order: Receiver<usize>,
+    output_receivers: Vec<Receiver<R>>,
+
+    /// Lookup table to register worker thread number corresponding to tasks
+    task_registry: TaskRegistry,
+
     /// handles to join worker threads
     worker_thread: Option<Vec<JoinHandle<()>>>,
-    /// flag to stop workers from fetching new tasks
+
+    /// atomic flag to stop workers from fetching new tasks
     iterator_stopper: Arc<AtomicBool>,
-    /// indicate that workers have all been killed
+
+    /// if this is `true`, it must guarantee that all worker threads have stopped
     is_killed: bool,
+
+    /// current task id
+    current: usize,
 }
 
 impl<R> ParIterSync<R>
-where
-    R: Send + 'static,
+    where
+        R: Send + 'static,
 {
     ///
     /// the worker threads are dispatched in this `new` constructor!
     ///
     pub fn new<T, TL, F>(tasks: TL, task_executor: F) -> Self
-    where
-        F: Send + Clone + 'static + Fn(T) -> Result<R, ()>,
-        T: Send,
-        TL: Send + IntoIterator<Item = T>,
-        <TL as IntoIterator>::IntoIter: Send + 'static,
+        where
+            F: Send + Clone + 'static + Fn(T) -> Result<R, ()>,
+            T: Send + 'static,
+            TL: Send + IntoIterator<Item = T> + 'static,
+            <TL as IntoIterator>::IntoIter: Send + 'static,
     {
         let cpus = num_cpus::get();
         let iterator_stopper = Arc::new(AtomicBool::new(false));
-        // worker master
-        let (task_register, task_order) = unbounded();
-        let tasks = Arc::new(Mutex::new(tasks.into_iter()));
-        let mut handles = Vec::with_capacity(cpus);
-        let mut receivers = Vec::with_capacity(cpus);
-        for thread_number in 0..cpus {
-            let (sender, receiver) = bounded(MAX_SIZE_FOR_THREAD);
-            let task = tasks.clone();
-            let register = task_register.clone();
+
+        // `(1 + MAX_SIZE_FOR_THREAD)` * cpus as there might be one more fetching after send blocking
+        let task_registry: TaskRegistry = TaskRegistry::new((1 + MAX_SIZE_FOR_THREAD) * cpus);
+
+        // this thread dispatches tasks to worker threads
+        let (dispatcher, task_receiver) = bounded(BUFFER_SIZE);
+        let sender_thread = thread::spawn(move || {
+            for (task_id, t) in tasks.into_iter().enumerate() {
+                if dispatcher.send((t, task_id)).is_err() {
+                    break;
+                }
+            }
+        });
+
+        // spawn worker threads
+        let mut handles = Vec::with_capacity(cpus + 1);
+        let mut output_receivers = Vec::with_capacity(cpus);
+        for thread_number in 0..cpus as isize {
+            let (output_sender, output_receiver) = bounded(MAX_SIZE_FOR_THREAD);
+            let task_receiver = task_receiver.clone();
+            let task_registry = task_registry.to_write();
             let iterator_stopper = iterator_stopper.clone();
             let task_executor = task_executor.clone();
 
             // workers
             let handle = thread::spawn(move || {
                 loop {
+                    // check stopper flag
                     if iterator_stopper.load(Ordering::SeqCst) {
                         break;
                     }
-                    match get_task(&task, &register, thread_number) {
-                        // finish
+                    //  fetch task and register thread number
+                    match get_task(&task_receiver, &task_registry, thread_number) {
+                        // stop if no more task
                         None => break,
                         Some(task) => match task_executor(task) {
                             Ok(blk) => {
-                                sender.send(blk).unwrap();
+                                // send output
+                                output_sender.send(blk).unwrap();
                             }
                             Err(_) => {
+                                // stop other thread when Error is returned
                                 iterator_stopper.fetch_or(true, Ordering::SeqCst);
                                 break;
                             }
@@ -346,36 +526,34 @@ where
                     }
                 }
             });
-            receivers.push(receiver);
+            output_receivers.push(output_receiver);
             handles.push(handle);
         }
+        handles.push(sender_thread);
 
         ParIterSync {
-            receivers,
-            task_order,
+            output_receivers,
+            task_registry,
             worker_thread: Some(handles),
             iterator_stopper,
             is_killed: false,
+            current: 0,
         }
     }
 }
 
 impl<R> ParIterSync<R> {
     ///
-    /// stop workers from fetching new tasks, and flush remaining works
-    /// to prevent blocking.
+    /// - stop workers from fetching new tasks
+    /// - pull one result from each worker to prevent `send` blocking
     ///
     pub fn kill(&mut self) {
         if !self.is_killed {
             // stop threads from getting new tasks
             self.iterator_stopper.fetch_or(true, Ordering::SeqCst);
-            // flush the remaining tasks in the channel
-            loop {
-                let _ = match self.task_order.recv() {
-                    Ok(thread_number) => self.receivers.get(thread_number).unwrap().recv(),
-                    // all workers have stopped
-                    Err(_) => break,
-                };
+            // receive one for each channel to prevent blocking
+            for receiver in &self.output_receivers {
+                let _ = receiver.try_recv();
             }
             // loop break only when task_order is dropped (all workers have stopped)
             self.is_killed = true;
@@ -384,30 +562,29 @@ impl<R> ParIterSync<R> {
 }
 
 ///
-/// A helper function that locks tasks,
-/// register thread_number and task_number
-/// before releasing tasks lock.
+/// A helper function to receive task from task receiver.
+/// - it also registers thread ID into task registry immediately.
+///
+/// It guarantees to return None if and only if there is no more new task.
 ///
 #[inline(always)]
-fn get_task<T, TL>(
-    tasks: &Arc<Mutex<TL>>,
-    register: &Sender<usize>,
-    thread_number: usize,
+fn get_task<T>(
+    tasks: &Receiver<(T, usize)>,
+    registry: &TaskRegistryWrite,
+    thread_number: isize,
 ) -> Option<T>
-where
-    T: Send,
-    TL: Iterator<Item = T>,
+    where
+        T: Send,
 {
     // lock task list
-    let mut task = tasks.lock().unwrap();
-    let next_task = task.next();
-    // register task stealing
-    match next_task {
-        Some(task) => {
-            register.send(thread_number).unwrap();
+    // let mut task = tasks.lock().unwrap();
+    // registry task stealing
+    match tasks.recv() {
+        Ok((task, task_id)) => {
+            registry.register(task_id, thread_number);
             Some(task)
         }
-        None => None,
+        Err(_) => None,
     }
 }
 
@@ -421,10 +598,15 @@ impl<R> Iterator for ParIterSync<R> {
         if self.is_killed {
             return None;
         }
-        match self.task_order.recv() {
-            Ok(thread_number) => {
-                match self.receivers.get(thread_number).unwrap().recv() {
+
+        // look up which thread to fetch result from
+        match self.task_registry.lookup(self.current) {
+            // no more task
+            None => None,
+            Some(thread_num) => {
+                match self.output_receivers[thread_num as usize].recv() {
                     Ok(block) => {
+                        self.current += 1;
                         Some(block)
                     }
                     // some worker have stopped
@@ -434,8 +616,6 @@ impl<R> Iterator for ParIterSync<R> {
                     }
                 }
             }
-            // all workers have stopped
-            Err(_) => None,
         }
     }
 }
@@ -575,7 +755,9 @@ mod test_par_iter {
 
             let results: Vec<i32> = (0..resource_captured.len())
                 .into_par_iter_sync(move |a| error_at_1000(&resource_captured_1, a as i32))
-                .into_par_iter_sync(move |a| Ok(resource_captured.get(a as usize).unwrap().to_owned()))
+                .into_par_iter_sync(move |a| {
+                    Ok(resource_captured.get(a as usize).unwrap().to_owned())
+                })
                 .into_par_iter_sync(move |a| {
                     Ok(resource_captured_2.get(a as usize).unwrap().to_owned())
                 })
